package org.arefgard.icerya.aop;

import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.Modifier;

import org.apache.oro.text.perl.Perl5Util;
import org.arefgard.icerya.aop.annotation.AfterAdvice;
import org.arefgard.icerya.aop.annotation.BeforeAdvice;
import org.arefgard.icerya.aop.annotation.DelegateAdvice;
import org.arefgard.icerya.aop.annotation.ThrowAdvice;

/**
 * AXyNgD荞܂ꂽvLVNX𐶐
 * t@NgB
 * 
 * @author Takashi Yamashina
 * @since 1.0.0
 * 
 */
public class AspectProxyFactory {
	
	private static final String FIELD_NAME = "_interceptorobject";
	
	private Class targetClass;
	Map<List<String>, Object> interceptors;
	
	Map<CtMethod, List<Method>> interceptMethods = new java.util.HashMap<CtMethod, List<Method>>();
	
	/** BeforeAdvice */
	private Map<String, List<AdviceDef>> beforeAdvices = new java.util.HashMap<String, List<AdviceDef>>();
	/** AfterAdvice */
	private Map<String, List<AdviceDef>> afterAdvices = new java.util.HashMap<String, List<AdviceDef>>();
	/** ThrowAdvice */
	private Map<String, List<AdviceDef>> throwAdvices = new java.util.HashMap<String, List<AdviceDef>>();
	/** DelegateAdvice */
	private Map<String, List<AdviceDef>> delegateAdvices = new java.util.HashMap<String, List<AdviceDef>>();
	
	public AspectProxyFactory(Class targetClass) {
		this.interceptors = new java.util.HashMap<List<String>, Object>();
		this.targetClass = targetClass;
	}
	
	public void addInterceptor(List<String> patterns, Object interceptor) {
		this.interceptors.put(patterns, interceptor);
	}
	
	public Class getProxy() throws AspectException {
		Class proxyClass = null;
		ClassPool pool = ClassPool.getDefault();
		try {
		
			// ^[QbgNXCtClass̎擾
			CtClass targetCtClass = pool.get(this.targetClass.getName());
			// vLVNX̐
			String targetPackageName = targetCtClass.getPackageName() + ".proxy.";
			String targetClassName = targetCtClass.getSimpleName() + "Proxy";
			CtClass ctClass = pool.makeClass(targetPackageName + targetClassName, targetCtClass);
			
			// C^[Zv^\bhƂɐ
			
			CtMethod[] ctMethods = targetCtClass.getDeclaredMethods();
			
			for(List<String> patterns : this.interceptors.keySet()) {
				Object interceptor = this.interceptors.get(patterns);
				
				for(CtMethod ctMethod : ctMethods) {
					if(match(patterns, ctMethod.getName())) {
						Method[] methods = interceptor.getClass().getDeclaredMethods();
						for(Method method : methods) {
							BeforeAdvice before = method.getAnnotation(BeforeAdvice.class);
							if(before != null) {
								List<AdviceDef> list = null;
								AdviceDef advice = new AdviceDef();
								advice.setClassName(interceptor.getClass().getSimpleName());
								advice.setMethodName(method.getName());
								Class[] argClasses = method.getParameterTypes();
								String[] argStrs = new String[argClasses.length];
								for(int i = 0; i < argClasses.length; i++) {
									argStrs[i] = argClasses[i].getName();
								}
								advice.setArgs(argStrs);
								
								if(beforeAdvices.containsKey(ctMethod)) {
									list = beforeAdvices.get(ctMethod);
								}else {
									list = new java.util.ArrayList<AdviceDef>();
								}
								list.add(advice);
								
								StringBuffer key = new StringBuffer();
								key.append(ctMethod.getName());
								for(CtClass cls : ctMethod.getParameterTypes()) {
									key.append(cls.getName());
								}
								beforeAdvices.put(key.toString(), list);
							}
							AfterAdvice after = method.getAnnotation(AfterAdvice.class);
							if(after != null) {
								List<AdviceDef> list = null;
								AdviceDef advice = new AdviceDef();
								advice.setClassName(interceptor.getClass().getSimpleName());
								advice.setMethodName(method.getName());
								Class[] argClasses = method.getParameterTypes();
								String[] argStrs = new String[argClasses.length];
								for(int i = 0; i < argClasses.length; i++) {
									argStrs[i] = argClasses[i].getName();
								}
								advice.setArgs(argStrs);
								
								if(afterAdvices.containsKey(ctMethod)) {
									list = afterAdvices.get(ctMethod);
								}else {
									list = new java.util.ArrayList<AdviceDef>();
								}
								list.add(advice);
								
								StringBuffer key = new StringBuffer();
								key.append(ctMethod.getName());
								for(CtClass cls : ctMethod.getParameterTypes()) {
									key.append(cls.getName());
								}
								afterAdvices.put(key.toString(), list);
							}
							ThrowAdvice throwAdvice = method.getAnnotation(ThrowAdvice.class);
							if(throwAdvice != null) {
								List<AdviceDef> list = null;
								AdviceDef advice = new AdviceDef();
								advice.setClassName(interceptor.getClass().getSimpleName());
								advice.setMethodName(method.getName());
								Class[] argClasses = method.getParameterTypes();
								String[] argStrs = new String[argClasses.length];
								for(int i = 0; i < argClasses.length; i++) {
									argStrs[i] = argClasses[i].getName();
								}
								advice.setArgs(argStrs);
								
								if(throwAdvices.containsKey(ctMethod)) {
									list = throwAdvices.get(ctMethod);
								}else {
									list = new java.util.ArrayList<AdviceDef>();
								}
								
								advice.setValues(throwAdvice.value());
								
								list.add(advice);
								StringBuffer key = new StringBuffer();
								key.append(ctMethod.getName());
								for(CtClass cls : ctMethod.getParameterTypes()) {
									key.append(cls.getName());
								}
								throwAdvices.put(key.toString(), list);
							}
							DelegateAdvice delegateAdvice = method.getAnnotation(DelegateAdvice.class);
							if(delegateAdvice != null) {
								List<AdviceDef> list = null;
								AdviceDef advice = new AdviceDef();
								advice.setClassName(interceptor.getClass().getSimpleName());
								advice.setMethodName(method.getName());
								Class[] argClasses = method.getParameterTypes();
								String[] argStrs = new String[argClasses.length];
								for(int i = 0; i < argClasses.length; i++) {
									argStrs[i] = argClasses[i].getName();
								}
								advice.setArgs(argStrs);
								
								if(delegateAdvices.containsKey(ctMethod)) {
									list = delegateAdvices.get(ctMethod);
								}else {
									list = new java.util.ArrayList<AdviceDef>();
								}
								list.add(advice);
								
								StringBuffer key = new StringBuffer();
								key.append(ctMethod.getName());
								for(CtClass cls : ctMethod.getParameterTypes()) {
									key.append(cls.getName());
								}
								delegateAdvices.put(key.toString(), list);
							}
						}
						put(ctMethod, interceptor);
					}
				}
			}
			
			// C^[Zv^NXɒ`B
			for(Object interceptor : this.interceptors.values()) {
				// tB[h̒ǉ
				CtField field = CtField.make(interceptor.getClass().getName() + " " + FIELD_NAME + interceptor.getClass().getSimpleName() + " = new " + interceptor.getClass().getName() + "();", ctClass);
				ctClass.addField(field);
			}

			// \bhƂɃvLV\bh`
			for(CtMethod method : this.interceptMethods.keySet()) {
				StringBuffer key = new StringBuffer();
				key.append(method.getName());
				for(CtClass cls : method.getParameterTypes()) {
					key.append(cls.getName());
				}
				CtMethod newMethod = createCtMethod(ctClass, method, beforeAdvices.get(key.toString()), afterAdvices.get(key.toString()), throwAdvices.get(key.toString()), delegateAdvices.get(key.toString()));
				ctClass.addMethod(newMethod);
			}
			
			// vLVNXClass𓾂
			proxyClass = ctClass.toClass();
			return proxyClass;
		}catch(Exception e) {
			throw new AspectException(e);
		}
	}
	
	/**
	 * TuNXɃ\bh`B
	 * @param ctClass
	 * @param ctMethod
	 * @throws Exception
	 */
	private CtMethod createCtMethod(CtClass ctClass,
									 CtMethod ctMethod, 
									 List<AdviceDef> beforeAdvices, 
									 List<AdviceDef> afterAdvices,
									 List<AdviceDef> throwAdvices,
									 List<AdviceDef> delegateAdvices) throws Exception {
		
		StringBuffer buf = new StringBuffer();
		// ANZXwq
		buf.append(Modifier.toString(ctMethod.getModifiers()));
		buf.append(" ");
		// ߂l
		buf.append(ctMethod.getReturnType().getName());
		buf.append(" ");
		// \bh
		buf.append(ctMethod.getName());
		buf.append("(");
		CtClass[] args = ctMethod.getParameterTypes();
		int length = args.length;
		for(int i = 0; i < length; i++) {
			buf.append(args[i].getName());
			buf.append(" ");
			buf.append("arg" + i);
			if(i != length - 1) {
				buf.append(",");
			}
		}
		buf.append(") {");
		
		// AfterAdvice܂ThrowAdviceꍇɂ́Atryݒ肷
		if((afterAdvices != null && afterAdvices.size() != 0)
				|| (throwAdvices != null && throwAdvices.size() != 0)) {
			buf.append("try {");
		}
		
		// Before AdviceD荞
		for(AdviceDef method : beforeAdvices) {
			buf.append("this." + FIELD_NAME + method.getClassName() + "." + method.getMethodName());
			buf.append("(");
			String[] methodArgs = method.getArgs();
			for(int i = 0; i < methodArgs.length; i++) {
				buf.append(methodArgs[i]);
				buf.append(" ");
				buf.append("arg" + i);
				if(i != length - 1) {
					buf.append(",");
				}
			}
			buf.append(");");
		}
		
		// DelegateAdvicȅꍇAeNX̃\bh̑ɁAfQ[g\bhĂяo
		if(delegateAdvices != null && delegateAdvices.size() != 0) {
			AdviceDef method = delegateAdvices.get(0);
			
			// e\bȟĂяo
			if(!ctMethod.getReturnType().getName().equals("void")) {
				// ߂l
				buf.append("return ");
			}
			buf.append("this." + FIELD_NAME + method.getClassName() + "." + method.getMethodName());
			buf.append(ctMethod.getName());
			buf.append("(");
			for(int i = 0; i < length; i++) {
				buf.append("arg" + i);
				if(i != length - 1) {
					buf.append(",");
				}
			}
			buf.append(");");
		}else {
			// e\bȟĂяo
			if(!ctMethod.getReturnType().getName().equals("void")) {
				// ߂l
				buf.append("return ");
			}
			buf.append("super.");
			buf.append(ctMethod.getName());
			buf.append("(");
			for(int i = 0; i < length; i++) {
				buf.append("arg" + i);
				if(i != length - 1) {
					buf.append(",");
				}
			}
			buf.append(");");
		}
		
		if((afterAdvices != null && afterAdvices.size() != 0)
				|| (throwAdvices != null && throwAdvices.size() != 0)) {
			buf.append("}");
		}
		
		// Throw AdviceD荞
		if(throwAdvices != null) {
			for(AdviceDef method : throwAdvices) {
				for(String ex : method.getValues()) {
					buf.append("catch(");
					buf.append(ex);
					buf.append(" e){");
					buf.append("this." + FIELD_NAME + method.getClassName() + "." + method.getMethodName());
					buf.append("(e);");
					buf.append("}");
				}
			}
		}
		
		// After AdviceD荞
		if(afterAdvices != null && afterAdvices.size() != 0) {
			buf.append("finally {");
		}
		if(afterAdvices != null) {
			for(AdviceDef method : afterAdvices) {
				buf.append("this." + FIELD_NAME + method.getClassName() + "." + method.getMethodName());
				buf.append("(");
				String[] methodArgs = method.getArgs();
				for(int i = 0; i < methodArgs.length; i++) {
					buf.append(methodArgs[i]);
					buf.append(" ");
					buf.append("arg" + i);
					if(i != length - 1) {
						buf.append(",");
					}
				}
				buf.append(");");
			}
		}
		if(afterAdvices != null && afterAdvices.size() != 0) {
			buf.append("}");
		}
		
		buf.append("}");
		
		return CtMethod.make(buf.toString(), ctClass);
	}
	
	private boolean match(List<String> patterns, String name) {
		Perl5Util perl = new Perl5Util();
		for(String pattern : patterns) {
			if(perl.match("/" + pattern + "/", name)) {
				return true;
			}
		}
		
		return false;
	}
	
	private void put(CtMethod method, Object obj) {
		List list = null;
		if(this.interceptMethods.containsKey(method)) {
			list = this.interceptMethods.get(method);
		}else {
			list = new java.util.ArrayList();
		}
		
		list.add(obj);
		this.interceptMethods.put(method, list);
	}
	
}
